from plone.formwidget.recurrence import _
from plone.formwidget.recurrence import pl_message
from zope.i18n import translate


def translations(request):
    translation_table = dict(
        displayUnactivate=translate(_("display_unactivate"), context=request),
        displayActivate=translate(_("display_activate"), context=request),
        add_rules=translate(_("add_rules"), context=request),
        edit_rules=translate(_("edit_rules"), context=request),
        delete_rules=translate(_("delete_rules"), context=request),
        add=translate(_("add"), context=request),
        refresh=translate(_("refresh"), context=request),
        title=translate(_("title"), context=request),
        preview=translate(_("preview"), context=request),
        addDate=translate(_("add_date"), context=request),
        recurrenceType=translate(_("recurrence_type"), context=request),
        dailyInterval1=translate(_("daily_interval_1"), context=request),
        dailyInterval2=translate(_("daily_interval_2"), context=request),
        weeklyInterval1=translate(_("weekly_interval_1"), context=request),
        weeklyInterval2=translate(_("weekly_interval_2"), context=request),
        weeklyWeekdays=translate(_("weekly_weekdays"), context=request),
        weeklyWeekdaysHuman=translate(_("weekly_weekdays_human"), context=request),
        monthlyInterval1=translate(_("monthly_interval_1"), context=request),
        monthlyInterval2=translate(_("monthly_interval_2"), context=request),
        monthlyDayOfMonth1=translate(_("monthly_day_of_month_1"), context=request),
        monthlyDayOfMonth1Human=translate(
            _("monthly_day_of_month_1_human"), context=request
        ),
        monthlyDayOfMonth2=translate(_("monthly_day_of_month_2"), context=request),
        monthlyDayOfMonth3=translate(_("monthly_day_of_month_3"), context=request),
        monthlyDayOfMonth4=translate(_("monthly_day_of_month_4"), context=request),
        monthlyWeekdayOfMonth1=translate(
            _("monthly_weekday_of_month_1"), context=request
        ),
        monthlyWeekdayOfMonth1Human=translate(
            _("monthly_weekday_of_month_1_human"), context=request
        ),
        monthlyWeekdayOfMonth2=translate(
            _("monthly_weekday_of_month_2"), context=request
        ),
        monthlyWeekdayOfMonth3=translate(
            _("monthly_weekday_of_month_3"), context=request
        ),
        monthlyRepeatOn=translate(_("monthly_repeat_on"), context=request),
        yearlyInterval1=translate(_("yearly_interval_1"), context=request),
        yearlyInterval2=translate(_("yearly_interval_2"), context=request),
        yearlyDayOfMonth1=translate(_("yearly_day_of_month_1"), context=request),
        yearlyDayOfMonth1Human=translate(
            _("yearly_day_of_month_1_human"), context=request
        ),
        yearlyDayOfMonth2=translate(_("yearly_day_of_month_2"), context=request),
        yearlyDayOfMonth3=translate(_("yearly_day_of_month_3"), context=request),
        yearlyWeekdayOfMonth1=translate(
            _("yearly_weekday_of_month_1"), context=request
        ),
        yearlyWeekdayOfMonth1Human=translate(
            _("yearly_weekday_of_month_1_human"), context=request
        ),
        yearlyWeekdayOfMonth2=translate(
            _("yearly_weekday_of_month_2"), context=request
        ),
        yearlyWeekdayOfMonth3=translate(
            _("yearly_weekday_of_month_3"), context=request
        ),
        yearlyWeekdayOfMonth4=translate(
            _("yearly_weekday_of_month_4"), context=request
        ),
        yearlyRepeatOn=translate(_("yearly_repeat_on"), context=request),
        range=translate(_("range"), context=request),
        rangeNoEnd=translate(_("range_no_end"), context=request),
        rangeByOccurrences1=translate(_("range_by_occurrences_1"), context=request),
        rangeByOccurrences1Human=translate(
            _("range_by_occurrences_1_human"), context=request
        ),
        rangeByOccurrences2=translate(_("range_by_occurrences_2"), context=request),
        rangeByEndDate=translate(_("range_by_end_date"), context=request),
        rangeByEndDateHuman=translate(_("range_by_end_date_human"), context=request),
        including=translate(_("including"), context=request),
        cancel=translate(_("cancel"), context=request),
        save=translate(_("save"), context=request),
        recurrenceStart=translate(_("recurrence_start"), context=request),
        additionalDate=translate(_("additional_date"), context=request),
        include=translate(_("include"), context=request),
        exclude=translate(_("exclude"), context=request),
        remove=translate(_("remove"), context=request),
        orderIndexes=[
            translate(_("order_indexes_first"), context=request),
            translate(_("order_indexes_second"), context=request),
            translate(_("order_indexes_third"), context=request),
            translate(_("order_indexes_fourth"), context=request),
            translate(_("order_indexes_last"), context=request),
        ],
        months=[
            translate(pl_message("month_jan"), context=request),
            translate(pl_message("month_feb"), context=request),
            translate(pl_message("month_mar"), context=request),
            translate(pl_message("month_apr"), context=request),
            translate(pl_message("month_may"), context=request),
            translate(pl_message("month_jun"), context=request),
            translate(pl_message("month_jul"), context=request),
            translate(pl_message("month_aug"), context=request),
            translate(pl_message("month_sep"), context=request),
            translate(pl_message("month_oct"), context=request),
            translate(pl_message("month_nov"), context=request),
            translate(pl_message("month_dec"), context=request),
        ],
        shortMonths=[
            translate(pl_message("month_jan_abbr"), context=request),
            translate(pl_message("month_feb_abbr"), context=request),
            translate(pl_message("month_mar_abbr"), context=request),
            translate(pl_message("month_apr_abbr"), context=request),
            translate(pl_message("month_may_abbr"), context=request),
            translate(pl_message("month_jun_abbr"), context=request),
            translate(pl_message("month_jul_abbr"), context=request),
            translate(pl_message("month_aug_abbr"), context=request),
            translate(pl_message("month_sep_abbr"), context=request),
            translate(pl_message("month_oct_abbr"), context=request),
            translate(pl_message("month_nov_abbr"), context=request),
            translate(pl_message("month_dec_abbr"), context=request),
        ],
        weekdays=[
            translate(pl_message("weekday_sun"), context=request),
            translate(pl_message("weekday_mon"), context=request),
            translate(pl_message("weekday_tue"), context=request),
            translate(pl_message("weekday_wed"), context=request),
            translate(pl_message("weekday_thu"), context=request),
            translate(pl_message("weekday_fri"), context=request),
            translate(pl_message("weekday_sat"), context=request),
        ],
        shortWeekdays=[
            translate(pl_message("weekday_sun_abbr"), context=request),
            translate(pl_message("weekday_mon_abbr"), context=request),
            translate(pl_message("weekday_tue_abbr"), context=request),
            translate(pl_message("weekday_wed_abbr"), context=request),
            translate(pl_message("weekday_thu_abbr"), context=request),
            translate(pl_message("weekday_fri_abbr"), context=request),
            translate(pl_message("weekday_sat_abbr"), context=request),
        ],
        longDateFormat=translate(_("long_date_format"), context=request),
        shortDateFormat=translate(
            pl_message("date_format_short_datepicker"), context=request
        ),
        unsupportedFeatures=translate(_("unsupported_features"), context=request),
        noTemplateMatch=translate(_("no_template_match"), context=request),
        multipleDayOfMonth=translate(_("multiple_day_of_month"), context=request),
        bysetpos=translate(_("bysetpos"), context=request),
        noRule=translate(_("no_rule"), context=request),
        noRepeatEvery=translate(_("no_repeat_every"), context=request),
        noEndDate=translate(_("no_end_date"), context=request),
        noRepeatOn=translate(_("no_repeat_on"), context=request),
        pastEndDate=translate(_("past_end_date"), context=request),
        noEndAfterNOccurrences=translate(
            _("no_end_after_n_occurrences"), context=request
        ),
        alreadyAdded=translate(_("already_added"), context=request),
        rtemplate=dict(
            daily=translate(_("template_daily"), context=request),
            mondayfriday=translate(_("template_mondayfriday"), context=request),
            weekdays=translate(_("template_weekdays"), context=request),
            weekly=translate(_("template_weekly"), context=request),
            monthly=translate(_("template_monthly"), context=request),
            yearly=translate(_("template_yearly"), context=request),
        ),
        # a bit wonky here, except is a reserved word
        **{"except": translate(_("except"), context=request)}
    )
    return translation_table



from dateutil import rrule
from plone.base.i18nl10n import _interp_regex
from plone.base.i18nl10n import datetime_formatvariables
from plone.base.i18nl10n import monthname_msgid
from plone.base.i18nl10n import monthname_msgid_abbr
from plone.base.i18nl10n import name_formatvariables
from plone.base.i18nl10n import weekdayname_msgid
from plone.base.i18nl10n import weekdayname_msgid_abbr
from Products.Five import BrowserView
from zope.i18n import interpolate
from zope.i18n import translate

import datetime
import json
import re


BATCH_DELTA = 3  # How many batches to show before + after current batch
BATCH_SIZE = 10  # How many items per batch

# Translations from dateinput formatting to Plone translation strings
# See http://flowplayer.org/tools/dateinput/index.html
DATEFORMAT_XLATE = [
    (re.compile(pattern), replacement)
    for (pattern, replacement) in (
        ("dddd", "${A}"),
        ("ddd", "${a}"),
        ("dd", "${d}"),
        ("!%d", "${e}"),  # Will include a leading space for 1-9
        ("mmmm", "${B}"),
        ("mmm", "${b}"),
        ("mm", "${m}"),
        ("!%m", "${m}"),  # Will include leading zero
        ("yyyy", "${Y}"),
        ("yy", "${y}"),
    )
]


def dateformat_xlate(dateformat):
    for regexp, replacement in DATEFORMAT_XLATE:
        dateformat = regexp.sub(replacement, dateformat)
    return dateformat


class RecurrenceView(BrowserView):
    def __call__(self):
        req = self.request
        req.response.setHeader("Content-type", "application/json")
        json_string = self.json_string
        return json.dumps(json_string)

    @property
    def json_string(self):
        occurrences = []
        data = self.request.form
        # Check for required parameters:
        for x in ("year", "month", "day", "rrule", "format"):
            if x not in data:
                self.request.response.setStatus(400)  # bad request
                return {}

        # Translate from the js dateformat style to the i18n style
        date_format = dateformat_xlate(data["format"])
        start_date = datetime.datetime(
            int(data["year"]), int(data["month"]), int(data["day"])
        )
        rule = rrule.rrulestr(data["rrule"], dtstart=start_date)
        iterator = iter(rule)

        if "batch_size" in data:
            batch_size = int(data["batch_size"])
        else:
            batch_size = BATCH_SIZE

        if "start" in data:
            start = int(data["start"])
        else:
            start = 0

        cur_batch = start // batch_size
        start = cur_batch * batch_size  # Avoid stupid start-values

        if hasattr(rule, "_exdate"):
            exdates = sorted(rule._exdate)
        else:
            exdates = []

        # Loop through the start first dates, to skip them:
        i = 0
        occurrences = []
        while True:
            try:
                # Get a date
                date = next(iterator)
            except StopIteration:
                # No more dates
                break
            while exdates and date > exdates[0]:
                # There are exdates that appear before this date:
                if i < start:
                    # Skip them
                    exdates.pop(0)
                    i += 1
                else:
                    # include them
                    exdate = exdates.pop(0)
                    occurrences.append(
                        {
                            "date": exdate.strftime("%Y%m%dT%H%M%S"),
                            "formattedDate": self.date_format(exdate, date_format),
                            "type": "exdate",
                        }
                    )
                    i += 1

            if i >= batch_size + start:
                break  # We are done!

            i += 1
            if i <= start:
                # We are still iterating up to the first event, so skip this:
                continue

            # Add it to the results
            if date in getattr(rule, "_rdate", []):
                occurrence_type = "rdate"
            elif date == start_date:
                occurrence_type = "start"
            else:
                occurrence_type = "rrule"
            occurrences.append(
                {
                    "date": date.strftime("%Y%m%dT%H%M%S"),
                    "formattedDate": self.date_format(date, date_format),
                    "type": occurrence_type,
                }
            )

        while exdates:
            # There are exdates that are after the end of the recurrence.
            # Excluding the last dates make no sense, as you can change the
            # range instead, but we need to support it anyway.
            exdate = exdates.pop(0)
            occurrences.append(
                {
                    "date": exdate.strftime("%Y%m%dT%H%M%S"),
                    "formattedDate": exdate.strftime(date_format),
                    "type": "exdate",
                }
            )

        # Calculate no of occurrences, but only to a max of three times
        # the batch size. This will support infinite recurrence in a
        # useable way, as there will always be more batches.
        first_batch = max(0, cur_batch - BATCH_DELTA)
        last_batch = max(BATCH_DELTA * 2, cur_batch + BATCH_DELTA)
        maxcount = (batch_size * last_batch) - start

        num_occurrences = 0
        while True:
            try:
                next(iterator)
                num_occurrences += 1
            except StopIteration:
                break
            if num_occurrences >= maxcount:
                break

        # Total number of occurrences:
        num_occurrences += batch_size + start

        max_batch = (num_occurrences - 1) // batch_size
        if last_batch > max_batch:
            last_batch = max_batch
            first_batch = max(0, max_batch - (BATCH_DELTA * 2))

        batches = [
            ((x * batch_size) + 1, (x + 1) * batch_size)
            for x in range(first_batch, last_batch + 1)
        ]
        # cur_batch must be set relative to first_batch as the batches array
        # is cut down, but the cur_batch variable is not. With this we prevent
        # a list index out of range error
        cur_batch = cur_batch - first_batch
        # correct number of occurrences in current batch
        (cur_batch_start, cur_batch_end) = batches[cur_batch]
        if cur_batch_end > num_occurrences:
            batches[cur_batch] = (cur_batch_start, num_occurrences)

        batch_data = {
            "start": start,
            "end": num_occurrences,
            "batch_size": batch_size,
            "batches": batches,
            "currentBatch": cur_batch,
        }

        result = {"occurrences": occurrences, "batch": batch_data}
        return result

    def date_format(self, time, formatstring):
        # This is a simplified version of
        # Products.CMFPlone.i18nl10n.ulocalized_time
        # that can take any format string.

        # ${a}        Locale's abbreviated weekday name.
        # ${A}        Locale's full weekday name.
        # ${b}        Locale's abbreviated month name.
        # ${B}        Locale's full month name.
        # ${d}        Day of the month as a decimal number [01,31].
        # ${H}        Hour (24-hour clock) as a decimal number [00,23].
        # ${I}        Hour (12-hour clock) as a decimal number [01,12].
        # ${m}        Month as a decimal number [01,12].
        # ${M}        Minute as a decimal number [00,59].
        # ${p}        Locale's equivalent of either AM or PM.
        # ${S}        Second as a decimal number [00,61].
        # ${y}        Year without century as a decimal number [00,99].
        # ${Y}        Year with century as a decimal number.
        # ${Z}        Time zone name (no characters if no time zone exists).

        # get the format elements used in the formatstring
        mapping = {}
        formatelements = _interp_regex.findall(formatstring)
        # reformat the ${foo} to foo
        formatelements = [el[2:-1] for el in formatelements]

        # add used elements to mapping
        elements = [e for e in formatelements if e in datetime_formatvariables]

        # add weekday name, abbr. weekday name, month name, abbr month name
        week_included = True
        month_included = True

        name_elements = [e for e in formatelements if e in name_formatvariables]
        if not ("a" in name_elements or "A" in name_elements):
            week_included = False
        if not ("b" in name_elements or "B" in name_elements):
            month_included = False

        for key in elements:
            mapping[key] = time.strftime("%" + key)

        if week_included:
            weekday = int(time.strftime("%w"))  # weekday, sunday = 0
            if "a" in name_elements:
                mapping["a"] = weekdayname_msgid_abbr(weekday)
            if "A" in name_elements:
                mapping["A"] = weekdayname_msgid(weekday)
        if month_included:
            monthday = int(time.strftime("%m"))  # month, january = 1
            if "b" in name_elements:
                mapping["b"] = monthname_msgid_abbr(monthday)
            if "B" in name_elements:
                mapping["B"] = monthname_msgid(monthday)

        # translate translateable elements
        for key in name_elements:
            mapping[key] = translate(
                mapping[key], "plonelocales", context=self.request, default=mapping[key]
            )

        # Apply the data to the format string:
        return interpolate(formatstring, mapping)



from setuptools import find_packages
from setuptools import setup


version = "3.0.2"

setup(
    name="plone.formwidget.recurrence",
    version=version,
    description="Recurrence widget for Plone",
    long_description="%s\n%s" % (open("README.rst").read(), open("CHANGES.rst").read()),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Plone",
        "Framework :: Plone :: 6.0",
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="Plone Event Recurrence Date Time Widget z3c.form",
    author="Plone Foundation",
    author_email="plone-developers@lists.sourceforge.net",
    url="https://github.com/plone/plone.formwidget.recurrence",
    license="GPL",
    packages=find_packages(),
    namespace_packages=["plone", "plone.formwidget"],
    include_package_data=True,
    python_requires=">=3.8",
    zip_safe=False,
    install_requires=[
        "setuptools",
        "plone.base",
        "Products.CMFCore",
        "python-dateutil",
        "zope.component",
        "zope.i18n",
        "zope.i18nmessageid",
        "zope.interface",
    ],
    extras_require=dict(
        z3cform=[
            "z3c.form",
            "zope.schema",
            "zope.traversing",
        ],
        test=[
            "plone.app.testing",
            "plone.testing",
            "plone.app.z3cform",
        ],
    ),
    entry_points="""
    # -*- Entry points: -*-
    [z3c.autoinclude.plugin]
    target = plone
    """,
)
